#include "eel.h"        /* standard definitions */

/*
 * This is an implementation of snippets from Textmate and other editors
 * It actually uses a converted snippet file from those editors.
 * The snippet file is simply a text file with the snippet abbreviation
 * at the start of the line followed by a TAB, then the actual snippet which will replace
 * that abbreviation, terminated by $end. (Much like the templates extension in epsilon).
 * The main difference is that fields can be defined within the snippet that will be
 * highlighted when the snippet is inserted, and the cursor is placed at the start of the first one.
 * TAB will select the next field, as the fields are visited via TAB they are selected so you can
 * simply type to replace them.
 * 
 * Fields are specified within a snippet the same way as in textmate, ${n:prompt}, where n is a number
 * and defines the order in which the fields are visited via the tab. The prompt part is what is
 * displayed and highlighted initially, and serves to document what goes there. Also a simple tab stop can be
 * specified with $1, $2 etc, this server as a TAB stop, and is not highlighed.
 *
 * Fields can be mirrored by using the same number, in this case when you type text into the first one the second
 * one will be updated with the same text, the second one does not become a TAB stop.
 * 
 * The snippet is terminated by the exact string $end, so the snippet can span many lines.
 * Adding snippets is simply a matter of editing the snippet file in the .epsilon folder.
 * The naming of the file is the major mode (ie c, html, etc) Followed by _snippets.txt.
 * (eg ruby_snippets.txt). In the case of the c major mode the extension of the file being edited
 * is looked at and the name of the snippet file is taken from that, (eg for a .java file java_snippets.txt
 * is used). Right now only c, c++, cpp and java are recognised, this can be changed in the open_snippet_file()
 * routine below.
 *
 * This package is delivered with snippets for ruby/rails both .rb files and .rhtml.
 * Also some java snippets are provided.
 *
 * A Ruby script is provided which generates the snippets.txt files from the textmate bundles.
 *
 * INSTALL
 * -------
 *
 * Simply copy snippets.e and all the *_snippets.txt files to your .epsilon directory, add
 * (load-eel-from-path "snippets.e" 2) to your einit.ecm file.
 * Lastly bind the command insert_snippet to a good key combination (I use alt-space).
 * ~reg-tab "A- ": insert_snippet
 *
 * If you are doing ruby/rails development add this to your einit.ecm file to make rhtml files treated like html
 * and use the ruby_mode.e extension.
 * suffix-rhtml: macro "<!suffix-html>"
 * 
 * 
 */


/*
 * TODO
 *  Handle recursive tabstops
 *  Moving around with arrow keys while in tabstop mode selects stuff
 *  Have a macro capability, ie pass the selection to a program, and paste the results
 *  If the snippet file does not exist just use general_snippets.txt
 *  Expand MACROS like TM_FILENAME and TM_SELECTED_TEXT that are in some snippets
 *  Add a list snippets command, and completion for a snippet if a close match is found
 */

/*
 * read the snippets file into a buffer
 * If the Major mode is not C then try to open a snippet file
 * of majormode_snippets.txt
 * If it is c mode look at the file extension and open ext_snippets.txt
 *
 */

int open_snippet_file()
{
	int i;
	char *path;
	char bufn[128], fn[FNAMELEN];
	char mode[256];
	strcpy(mode, major_mode);
	if(any_uppercase(mode)){
		for(i=0;i<strlen(mode);i++){
			mode[i]= tolower(mode[i]);
		}
	}
	if(strcmp("c", mode) == 0){
		// may need to add modes here
		char *ext= get_extension(filename);
		if(strfcmp(ext, ".java") == 0){
			strcpy(mode, "java");
		}else if(strfcmp(ext, ".cpp") == 0 || strfcmp(ext, ".c++") == 0){
			strcpy(mode, "cpp");
		}else strcpy(mode, "c");
	}
	sprintf(bufn, "-%s-snippets", mode);
	sprintf(fn, "%s_snippets.txt", mode);
	
	if(!exist(bufn)){
		create(bufn);
		path= lookpath(fn);
		if ( !path ) {
			delete_buffer(bufn);
			error ( "Can't find file %s", fn);
		}
		bufname= bufn;
		i = err_file_read(path, FILETYPE_AUTO);
		if (i) {			/* error reading file */
			delete_buffer(bufn);
			error ( "Can't open file %s", path);
		}
	}else{
		bufname= bufn;
		point= 0;
	}
	return 1;
}

// this structure holds the tab stops
typedef struct {
	int order;
	int mirror;
	int has_prompt;
	int st, en;
	spot s1, s2;
} t_tabstop;

#define MAXTABSTOPS 32
// note max number of tab stops is 32 within a single snippet
typedef struct {
	int snippet;
	int ntabs;
	t_tabstop tabstop_list[MAXTABSTOPS];
	spot s1, s2;
	int curtabstop;
	int active;
	void (*prev_tab_cmd)();
} t_snippet;

// stores the current snippet for this buffer
zeroed buffer t_snippet cur_snippet;

// see if the tabstop already exists
int tabstop_exists(t_tabstop *tl, int size, int n)
{
	int i;
	for(i=0;i<size;i++){
		if(tl[i].order == n)
			return(i);
	}
	return -1;
}

// Extracts the tab stop from the given snippet which is in its own buffer
// recursively called as the body of a tabstop could have embedded tabstops
int extract_tabstop(int snipbuf, t_tabstop *tl, int *ntabs)
{
	// there are two types of tab stop $1, $2 etc and ${1:prompt}, this regex handles both
	char buf1[6];
	char *pat2= "(%$(<digit>+))|(%${(<digit>+):([^}]+)})";
	int scnt= *ntabs;
	int startpos= 0;
	while(re_search(1, pat2)){
		int d1= find_group(1, 1);
		int d2= find_group(2, 1);
		int d3= find_group(3, 1);
		int d4= find_group(4, 1);
		int d5= find_group(5, 1);
		
		if(d1 != 0){ // it is a $n tab stop
			int p1= find_group(2, 1); // start of number
			int p2= find_group(2, 0); // end of number
			grab(p1, p2, buf1);
			int n= strtoi(buf1, 10);
			if(n == 0) n= 9999; // so it gets sorted to end
			t_tabstop *pt= &tl[scnt++];
			pt->order= n;
			// copy text leading up to this into snip buffer
			buf_xfer(snipbuf, startpos, p1-1);
			pt->st= pt->en= get_buf_point(snipbuf);
			point= p2;
			pt->has_prompt= 0;
			pt->mirror= -1;
		}else if(d3 != 0){
			int p1= find_group(4, 1); // start of number
			int p2= find_group(4, 0); // end of number
			int p3= find_group(5, 1); // start of body
			int p4= find_group(5, 0); // end of body
			grab(p1, p2, buf1);
			int n= strtoi(buf1, 10);
			if(n == 0) n= 9999; // so it gets sorted to end
			int mirror= tabstop_exists(tl, scnt, n);
			t_tabstop *pt= &tl[scnt++];
			pt->mirror= mirror;
			pt->order= n;
			// text leading up to tabstop
			buf_xfer(snipbuf, startpos, p1-2);
			pt->st= get_buf_point(snipbuf); // starting offset of tabstop
			//extract_tabstop(snipbuf, tl, &scnt); // recursively call to extract embedded tabstops
			buf_xfer(snipbuf, p3, p4); // copy body of tab stop into snippet buffer
			pt->en= get_buf_point(snipbuf);
			point= p4+1;
			pt->has_prompt= 1;
		}
		startpos= point;
	}
	*ntabs= scnt;
	point= startpos;
	return 0;
}

t_snippet* find_snippet(char *name)
{
	char *result;
	char pat[132];
	char *original_buffer= bufname;
	int oldpos= point;
	save_var matchstart, matchend;
	int i, j;

	open_snippet_file();

	sprintf(pat, "^(%s)<Tab>(<any>*)%%$end", name);
	if(!re_search(RE_FORWARD|RE_SHORTEST, pat)){
		error("snippet %s not found", name);
	}

	int p1= find_group(2, 1);
	int p2= find_group(2, 0);
	
	// copy snippet into its own buffer
	int tmpbuf= tmp_buf();
	buf_xfer(tmpbuf, p1, p2);
	bufnum= tmpbuf;
	
	// alocate buffer for resulting snippet
	int snipbuf= tmp_buf();

	// set to start of the snippet buffer
	point= 0;

	// find each tab stop, and build the snippet inserting the prompt
	// and remember the offsets for highlighing and tabbing too
	t_tabstop tabstop_list[MAXTABSTOPS];
	int scnt= 0;
	extract_tabstop(snipbuf, tabstop_list, &scnt);
	
	// copy the rest of the snippet
	buf_xfer(snipbuf, point, size()+1);

	bufname= original_buffer;
	point = oldpos;

	cur_snippet.ntabs= scnt;

	// copy the tabstops
	for(i=0;i<scnt;i++){
		cur_snippet.tabstop_list[i]= tabstop_list[i];
	}

	// bubble sort by order
	for (i=0; i<scnt-1; i++) {
		for (j=0; j<scnt-1-i; j++){
			if(cur_snippet.tabstop_list[j+1].order < cur_snippet.tabstop_list[j].order){
				t_tabstop tmp= cur_snippet.tabstop_list[j];
				cur_snippet.tabstop_list[j]= cur_snippet.tabstop_list[j+1];
				cur_snippet.tabstop_list[j+1]= tmp;
			}
		}
	}

	// store the buffer the snippet is in
	cur_snippet.snippet= snipbuf;

	buf_delete(tmpbuf);
	return &cur_snippet;
}

// command test_regex()
// {
// 	char *pat2= "(%$(<digit>+))|(%${(<digit>+):([^}]+)})";
// 
// 	while(re_search(1, pat2)){
// 		char snip[256];
// 		char snip_name[32];
// 		int p1= find_group(1, 1);
// 		int p2= find_group(2, 1);
// 		int p3= find_group(3, 1);
// 		int p4= find_group(4, 1);
// 		int p5= find_group(5, 1);
// 
// 		buffer_printf("test", "p1= %d, p2= %d, p3= %d, p4= %d, p5= %d\n", p1, p2, p3, p4, p5);
// 	}
// 
// }

int inside_snippet()
{
	if(point >= *cur_snippet.s1 && point < *cur_snippet.s2)
		return 1;

	return 0;
}

// place holder for v13 where we can turn off snippets of we click outside the snippet
// int check_if_left_snippet()
// {
// 	int i;
// 	if(!inside_snippet){
// 		deactivate_snippet();
// 		delete_buffer_when_displaying(bufnum, check_if_left_snippet);
// 	}
// }

int deactivate_snippet(); // forward reference

// This routine is used to mirror any tab stops, and check if we are still inside the snippet
int snippet_when_idle(int cnt)
{
	int i;
	if(cur_snippet.active == 0) return -1;

	// if we moved point out of snippet disable snippet
	if(!inside_snippet()){
		deactivate_snippet();
		return -1;
	}
	
	for(i=0;i<cur_snippet.ntabs;i++){
		if(cur_snippet.tabstop_list[i].mirror >= 0){
			// find the tabstop this is mirroring
			int m= tabstop_exists(cur_snippet.tabstop_list, cur_snippet.ntabs, cur_snippet.tabstop_list[i].order);
			if(m == -1) continue; // this should never happen
			say("idle: %d - mirror %d to %d", cnt, m, i);
			// get text from that tabstop and overwrite this text with that
			char *buf1= NULL, *buf2= NULL;
			grab_expanding(*cur_snippet.tabstop_list[m].s1, *cur_snippet.tabstop_list[m].s2, &buf1, 10);
			grab_expanding(*cur_snippet.tabstop_list[i].s1, *cur_snippet.tabstop_list[i].s2, &buf2, 10);
			if(strcmp(buf1, buf2) != 0){ // only if different
				save_spot point;
				save_spot mark;
				int p1= *cur_snippet.tabstop_list[i].s1;
				int p2= *cur_snippet.tabstop_list[i].s2;
				delete(p1, p2);
				point= *cur_snippet.tabstop_list[i].s1;
				stuff(buf1);
			}
			free(buf1);
			free(buf2);
		}
	}
	return -1;
}

// TAB function when inside a snippet, jumps to next snippet field
command snippet_tab_key()
{
	if (cur_snippet.active && inside_snippet())
		to_next_snippet_field();
	else if(cur_snippet.prev_tab_cmd)
		(*cur_snippet.prev_tab_cmd)();
}

// looks up the snippet abbreviation and inserts the snippet
#define IDENT_PATTERN	"[a-zA-Z0-9_\xC0-\xD6\xD8-\xF6\xF8-\xFF:]+" // Matches an identifier.
color_class ident_completion;
color_class snippet_tab_stop = color_class ident_completion;

// This is the main command that will ilook up a snippet and insert it
// it looks at the work to the left of point and looks it up, if there isn;lt one it will
// prompt for the snippet.
// bind it to what ever key you prefer.
command insert_snippet()
{
	char tag[32];
	int i;
	int tag_point= 0, tag_mark;
	
	// get the snippet to insert
	mark= point;
	
	if (parse_string(-1, IDENT_PATTERN, tag)){
		fix_region();
		tag_point= point;
		tag_mark= mark;
	}else
		get_string(tag, "Snippet: ");

	point= mark;
	
	// lookup the snippet and insert it
	find_snippet(tag);
	if(tag_point != 0)
		delete ( tag_point, tag_mark );		
	int svpt1= point;
	grab_buffer(cur_snippet.snippet);
	cur_snippet.s2= alloc_spot(1); // end of snippet
	point= svpt1;
	cur_snippet.s1= alloc_spot(1); // start of snippet
	
	// now set the spots for each field, s1 must be left inserting for now as we do an indent
	for(i=0;i<cur_snippet.ntabs;i++){
		cur_snippet.tabstop_list[i].s1= alloc_spot(1);
		set_spot(&cur_snippet.tabstop_list[i].s1, *cur_snippet.s1 + cur_snippet.tabstop_list[i].st);
		if(cur_snippet.tabstop_list[i].has_prompt){
			cur_snippet.tabstop_list[i].s2= alloc_spot(1);
			set_spot(&cur_snippet.tabstop_list[i].s2, *cur_snippet.s1 + cur_snippet.tabstop_list[i].en);
		}
	}

	// indent according to current indent rules
	point= *cur_snippet.s1;
	mark= *cur_snippet.s2;
	indent_region();

	// now we need to change the s1 spot of each TAB STOP to right insert, and highlight the tab stops
	for(i=0;i<cur_snippet.ntabs;i++){
		int saves1= *cur_snippet.tabstop_list[i].s1;
		free_spot(cur_snippet.tabstop_list[i].s1);
		cur_snippet.tabstop_list[i].s1= alloc_spot(0);
		set_spot(&cur_snippet.tabstop_list[i].s1, saves1);
		if(cur_snippet.tabstop_list[i].has_prompt){
			add_region(cur_snippet.tabstop_list[i].s1, cur_snippet.tabstop_list[i].s2,  color_class snippet_tab_stop, REGNORM, i+1000);
		}
	}
	

	// position cursor at first tab stop if there is one
	cur_snippet.curtabstop= 0;

	// set TAB to cycle through the tab stops
	if (mode_keys['\t'] != snippet_tab_key) {
		cur_snippet.prev_tab_cmd = mode_keys['\t'];
		mode_keys['\t'] = snippet_tab_key;
	}
	//add_buffer_when_displaying(bufnum, check_if_left_snippet);
	to_next_snippet_field();
	cur_snippet.active= 1;
	add_buffer_when_idle(bufnum, snippet_when_idle);
}

// deactivate snippet and release its resources
int deactivate_snippet()
{
	int i;
	cur_snippet.active= 0;
	highlight_off();	

	// release allocations
	buf_delete(cur_snippet.snippet);
	for(i=0;i<cur_snippet.ntabs;i++){
		if(cur_snippet.tabstop_list[i].has_prompt){
			remove_region(i+1000);
			free_spot(cur_snippet.tabstop_list[i].s2);
		}
		free_spot(cur_snippet.tabstop_list[i].s1);
	}
	cur_snippet.snippet= 0;
	free_spot(cur_snippet.s1);
	free_spot(cur_snippet.s2);

	// remove special TAB key command
	if (mode_keys['\t'] == snippet_tab_key && cur_snippet.prev_tab_cmd) {
		mode_keys['\t'] = cur_snippet.prev_tab_cmd;
		cur_snippet.prev_tab_cmd = 0;
	}
	delete_buffer_when_idle(bufnum, snippet_when_idle);
}   	

// TAB will go to the next parameter, until no more then it will go to the end of the line
// and disable the snippet and special TAB key function
void to_next_snippet_field()
{
	int i;
	int done= 0;
next:
	if(cur_snippet.curtabstop < cur_snippet.ntabs){
		// highlight the next tab stop
		int n= cur_snippet.curtabstop++;
		// skip mirrored tabstops
		if(cur_snippet.tabstop_list[n].mirror != -1)
			goto next;
		point= *cur_snippet.tabstop_list[n].s1;
		if(cur_snippet.tabstop_list[n].has_prompt){
			mark= *cur_snippet.tabstop_list[n].s2;
			//fix_region();
			set_region_type(REGNORM);
			highlight_on();
		}else{ // its a tabstop only
			mark= point;
			if ( !parse_string ( -1, "<^wspace>+<hspace>*", NULL ) ) {
				insert('\t'); // only if at begining of line
			}

			if(cur_snippet.tabstop_list[n].order == 9999){ // its the last stop leave point there
				done= 1;
			}
		}
	}else{
		// finished move point to end
		point= *cur_snippet.s2;
		mark= point;
		done= 1;
	}

	if(done){
		//delete_buffer_when_displaying(bufnum, check_if_left_snippet);
		deactivate_snippet();
	}
}

/*
 * Send the selected text to the given command, and return it as a string
 */
char *snippet_pipe_to_command(char *cmd)
{
	save_spot point, mark;
	char *inp_buffer= NULL;
	int oldbufnum= bufnum;
	
	say("");
	fix_region();

	// if a region is selected send it to the command
	if (is_highlight_on() && mark - point > 0){
		inp_buffer= temp_buf();
		xfer(inp_buffer, point, mark);
		highlight_off();
	}

	// run the show_columns ruby script which extracts columns for the given model in project path
	pipe_text(inp_buffer, "-snippet-pipe-output", cmd, NULL, PIPE_SYNCH|PIPE_CLEAR_BUF|PIPE_SKIP_SHELL , 0);

	// get the output into a string
	bufname= "-snippet-pipe-output";
	char *res= NULL;
	grab_expanding(0, size(), &res, 256);
	bufnum= oldbufnum;
	return res;
}

//command test_it()
//{
// 	char *res= snippet_pipe_to_command("pargs 1 2 3");
// 	stuff(res);
// 	free(res);

// 	if ( !parse_string ( -1, "<^wspace>+<hspace>*", NULL ) ) {
//  		 say("insert tab");
//  	 }else{
//  		 say("no");
//	}
//}
